@@ -200,8 +200,9 @@ get_default_keyring_file_for_name (const char *keyring_name)
 		}
 		
 		path = g_build_filename (dir, filename, NULL);
-				
 		g_free (filename);
+
+		version++;
 	} while (g_file_test (path, G_FILE_TEST_EXISTS));
 
 	g_free (base);
@@ -370,9 +371,11 @@ remove_keyring_file_from_disk (GnomeKeyring *keyring)
 {
 	int res;
 
-	g_assert (keyring->file != NULL);
+	/* Cannot remove session or memory based keyring */
+	if (keyring->file == NULL)
+		return FALSE;
+
 	res = unlink (keyring->file);
-		
 	return (res == 0);
 }
 
@@ -429,6 +432,9 @@ save_keyring_to_disk (GnomeKeyring *keyring)
 		if (fd != -1) {
 			fchmod (fd, S_IRUSR | S_IWUSR);
 			if (write_all (fd, out->str, out->len) == 0) {
+#ifdef HAVE_FSYNC
+			fsync (fd);
+#endif
 				close (fd);
 				if (rename (template, keyring->file) == 0) {
 					if (stat (keyring->file, &statbuf) == 0) {
@@ -823,6 +829,9 @@ set_default_keyring (GnomeKeyring *keyring)
 		if (keyring != NULL && keyring->keyring_name != NULL) {
 			write_all (fd, keyring->keyring_name,
 				   strlen (keyring->keyring_name));
+#ifdef HAVE_FSYNC
+			fsync (fd);
+#endif
 		}
 		close (fd);
 	}
